#include <pthread.h>
#include "dronedge.h"

static JSClassID jsDroneClassId;

static UA_Boolean pubsubShouldRun = true;
static UA_Boolean pubsubExited = false;

static UA_UInt32 nbDrone;
static UA_UInt32 nbSubscriber;
static JSValueConst *droneObjectIdList;
static MessageQueue messageQueue = {
    .head = NULL,
    .tail = NULL,
};
UA_String currentMessage;

pthread_mutex_t mutex;
pthread_cond_t threadCond;

// Drone class functions

static void js_drone_finalizer(JSRuntime *rt, JSValue val)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(val, jsDroneClassId);
    js_free_rt(rt, s);
}

static JSValue js_drone_ctor(JSContext *ctx, JSValueConst newTarget, 
                             int argc, JSValueConst *argv)
{
    JSDroneData *s;
    JSValue obj = JS_UNDEFINED;
    JSValue proto;
    uint32_t uint32;
    
    s = (JSDroneData *) js_mallocz(ctx, sizeof(*s));
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToUint32(ctx, &uint32, argv[0]))
        goto fail;
    s->id = (uint16_t)uint32;

    proto = JS_GetPropertyStr(ctx, newTarget, "prototype");
    if (JS_IsException(proto))
        goto fail;
    obj = JS_NewObjectProtoClass(ctx, proto, jsDroneClassId);
    JS_FreeValue(ctx, proto);
    if (JS_IsException(obj))
        goto fail;
    JS_SetOpaque(obj, s);
    return obj;
 fail:
    js_free(ctx, s);
    JS_FreeValue(ctx, obj);
    return JS_EXCEPTION;
}

static JSValue js_drone_init(JSContext *ctx, JSValueConst thisVal,
                             int argc, JSValueConst *argv)
{
    int nb;
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, thisVal, jsDroneClassId);
    if (!s)
        return JS_EXCEPTION;
    if (JS_ToInt32(ctx, &nb, argv[0]))
        return JS_EXCEPTION;
    droneObjectIdList[nb] = thisVal;
    return JS_UNDEFINED;
}

static JSValue js_drone_get(JSContext *ctx, JSValueConst thisVal, int magic)
{
    JSDroneData *s = (JSDroneData *) JS_GetOpaque2(ctx, thisVal, jsDroneClassId);
    JSValue res;
    if (!s)
        return JS_EXCEPTION;
    switch(magic) {
        case 0:
            return JS_NewUint32(ctx, s->id);
        case 1:
            return JS_NewFloat64(ctx, s->latitude);
        case 2:
            return JS_NewFloat64(ctx, s->longitude);
        case 3:
            return JS_NewFloat64(ctx, s->altitudeAbs);
        case 4:
            return JS_NewFloat64(ctx, s->altitudeRel);
        case 5:
            return JS_NewFloat64(ctx, s->yaw);
        case 6:
            return JS_NewFloat64(ctx, s->speed);
        case 7:
            return JS_NewFloat64(ctx, s->climbRate);
        case 8:
            pthread_mutex_lock(&mutex);
            res = JS_NewString(ctx, s->message);
            strcpy(s->message, "");
            pthread_cond_signal(&threadCond);
            pthread_mutex_unlock(&mutex);
            return res;
        default:
            return JS_EXCEPTION;
    }
}

static JSValue js_drone_set_message(JSContext *ctx, JSValueConst thisVal,
                                    int argc, JSValueConst *argv)
{
    struct messageNode *newNode;
    const char *message;

    message = JS_ToCString(ctx, argv[0]);
    if (strlen(message) > MAX_MESSAGE_SIZE) {
        UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Message too long");
        return JS_EXCEPTION;
    }

    newNode = (struct messageNode*)malloc(sizeof(struct messageNode));
    newNode->message = strdup(message);
    newNode->next = NULL;

    if (messageQueue.tail == NULL) {
        messageQueue.head = messageQueue.tail = newNode;
    } else {
        messageQueue.tail->next = newNode;
        messageQueue.tail = newNode;
    }

    JS_FreeCString(ctx, message);
    return JS_UNDEFINED;
}

static void delete_message_node(struct messageNode *node) {
    free(node->message);
    free(node);
}

static UA_Boolean UA_String_isEmpty(const UA_String *s) {
    return (s->length == 0 || s->data == NULL);
}

static void clear_message(UA_String message) {
    if (!UA_String_isEmpty(&message))
        UA_String_clear(&message); 
}

UA_String get_message(void)
{
    struct messageNode *current;

    current = messageQueue.head;
    if (current != NULL) {
        clear_message(currentMessage);
        currentMessage = UA_STRING_ALLOC(current->message);
        messageQueue.head = current->next == NULL ? (messageQueue.tail = NULL) : current->next;
        delete_message_node(current);
    }
    return currentMessage;
}

static JSClassDef jsDroneClass = {
    "Drone",
    .finalizer = js_drone_finalizer,
};

static const JSCFunctionListEntry js_drone_proto_funcs[] = {
    JS_CGETSET_MAGIC_DEF("id", js_drone_get, NULL, 0),
    JS_CGETSET_MAGIC_DEF("latitude", js_drone_get, NULL, 1),
    JS_CGETSET_MAGIC_DEF("longitude", js_drone_get, NULL, 2),
    JS_CGETSET_MAGIC_DEF("altitudeAbs", js_drone_get, NULL, 3),
    JS_CGETSET_MAGIC_DEF("altitudeRel", js_drone_get, NULL, 4),
    JS_CGETSET_MAGIC_DEF("yaw", js_drone_get, NULL, 5),
    JS_CGETSET_MAGIC_DEF("speed", js_drone_get, NULL, 6),
    JS_CGETSET_MAGIC_DEF("climbRate", js_drone_get, NULL, 7),
    JS_CGETSET_MAGIC_DEF("message", js_drone_get, NULL, 8),
    JS_CFUNC_DEF("init", 1, js_drone_init),
};

// pubsub functions

UA_UInt16 get_drone_id(UA_UInt32 nb) {
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[nb], jsDroneClassId);
    return s->id;
}

VariableData pubsub_get_value(UA_String identifier) {
    VariableData varDetails;
    UA_DataType type;
    UA_Double *array;

    for(UA_UInt32 i = 0; i < droneVariables.nbVariable; i++) {
        UA_String name = UA_STRING(droneVariables.variableArray[i].name);
        if(UA_String_equal(&identifier, &name)) {
            varDetails = droneVariables.variableArray[i];
            switch(varDetails.valueRank) {
                case UA_VALUERANK_SCALAR:
                    switch(varDetails.type) {
                        case UA_TYPES_STRING:
                            *(UA_String*)varDetails.value = varDetails.getter.getString();
                            break;
                        default:
                            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "UA_TYPE not handled");
                            break;
                    }
                    break;

                case UA_VALUERANK_ONE_DIMENSION:
                    type = UA_TYPES[varDetails.type];
                    size_t size = varDetails.arrayDimensions[0];
                    array = (UA_Double *) varDetails.getter.getArray();

                    if(type.pointerFree) {
                        memcpy(varDetails.value, array, type.memSize * size);
                    } else {
                        uintptr_t ptrs = (uintptr_t)array;
                        uintptr_t ptrd = (uintptr_t)varDetails.value;
                        UA_StatusCode retval = UA_STATUSCODE_GOOD;
                        for(size_t j = 0; j < size; ++j) {
                            retval |= UA_copy((void*)ptrs, (void*)ptrd, &type);
                            ptrs += type.memSize;
                            ptrd += type.memSize;
                        }
                        if(retval != UA_STATUSCODE_GOOD)
                            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Failed to update array");
                    }

                    free(array);
                    break;

                default:
                    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "UA_VALUERANK not handled");
                    break;
            }
        }
    }

    return varDetails;
}

void init_drone_node_id(UA_UInt32 id, UA_UInt32 nb, UA_UInt32 magic) {
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[nb], jsDroneClassId);
    switch(magic) {
        case 0:
            s->positionArrayId = id;
            break;
        case 1:
            s->speedArrayId = id;
            break;
        case 2:
            s->messageId = id;
            break;
        default:
            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_USERLAND, "Unknown variable id");
            break;
    }
}

void init_subscriber_node_id(UA_UInt32 id, UA_UInt32 nb, UA_UInt32 magic) {
    JSDroneData *s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[nb], jsDroneClassId);
    switch(magic) {
        case 0:
            s->messageId = id;
            break;
        default:
            UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_USERLAND, "Unknown variable id");
            break;
    }
}

static void pubsub_update_variables(UA_UInt32 id, const UA_DataValue *var, bool print)
{
    JSDroneData*  s;
    UA_String     uaStr;
    UA_Double*    positionArray;
    UA_Float*     speedArray;

    for(UA_UInt32 i = 0; i < nbDrone + nbSubscriber; i++) {
        s = (JSDroneData *) JS_GetOpaque(droneObjectIdList[i], jsDroneClassId);
        if (s->positionArrayId == id) {
            positionArray = (UA_Double*) var->value.data;
            s->latitude = positionArray[0];
            s->longitude = positionArray[1];
            s->altitudeAbs = positionArray[2];
            s->altitudeRel = positionArray[3];

            if (print) {
                UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT,
                  "Received position of drone %d: %f° %f° %fm %fm",
                  s->id, s->latitude, s->longitude, s->altitudeAbs, s->altitudeRel);
            }
            return;
        } else if (s->speedArrayId == id) {
            speedArray = (UA_Float*) var->value.data;
            s->yaw = speedArray[0];
            s->speed = speedArray[1];
            s->climbRate = speedArray[2];

            if (print) {
                UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT,
                  "Received speed of drone %d: %f° %fm/s %fm/s",
                  s->id, s->yaw, s->speed, s->climbRate);
            }
            return;
        } else if (s->messageId == id) {
            uaStr = *(UA_String*) var->value.data;

            pthread_mutex_lock(&mutex);
            while(strlen(s->message) != 0)
                pthread_cond_wait(&threadCond, &mutex);

            memcpy(s->message, uaStr.data, uaStr.length);
            s->message[uaStr.length] = '\0';

            pthread_mutex_unlock(&mutex);
            return;
        }
    }
    UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_CLIENT, "NodeId not found");
}

/*
 * arg 0 (string): IPv6 of the multicast group
 * arg 1 (string): port used for multicast communication
 * arg 2 (string): network interface used for multicast communication
 * arg 3 (int): ID of the drone
 * arg 4 (double): publication/subscription interval in ms
 * arg 5 (bool): true if there will be data to publish
*/
static JSValue js_run_pubsub(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    const char *ipv6;
    const char *port;
    const char *netIface;
    char *notConstNetIface;
    char urlBuffer[44];
    UA_UInt32 id;
    UA_Duration interval;
    bool isADrone;
    VariableStruct variables;
    InstanceData *instanceArray;
    UA_UInt32 nbPeer = nbDrone + nbSubscriber;
    int res;

    ipv6 = JS_ToCString(ctx, argv[0]);
    port = JS_ToCString(ctx, argv[1]);
    netIface = JS_ToCString(ctx, argv[2]);
    UA_snprintf(urlBuffer, sizeof(urlBuffer), "opc.udp://[%s]:%s/", ipv6, port);

    UA_String transportProfile =
        UA_STRING("http://opcfoundation.org/UA-Profile/Transport/pubsub-udp-uadp");

    UA_NetworkAddressUrlDataType networkAddressUrl;
    notConstNetIface = strdup(netIface);
    networkAddressUrl.networkInterface = UA_STRING(notConstNetIface);
    networkAddressUrl.url              = UA_STRING(urlBuffer);

    if (JS_ToUint32(ctx, &id, argv[3]))
        return JS_EXCEPTION;

    if (JS_ToFloat64(ctx, &interval, argv[4]))
        return JS_EXCEPTION;

    isADrone = JS_ToBool(ctx, argv[5]);
    variables = isADrone ? droneVariables : subscriberVariables;

    instanceArray = (InstanceData *) malloc((nbPeer) * sizeof(InstanceData));
    for(UA_UInt32 i = 0; i < nbDrone; i++) {
      instanceArray[i].variables = droneVariables;
      instanceArray[i].init_node_id = init_drone_node_id;
    }
    for(UA_UInt32 i = nbDrone; i < nbPeer; i++) {
      instanceArray[i].variables = subscriberVariables;
      instanceArray[i].init_node_id = init_subscriber_node_id;
    }

    res = runPubsub(&transportProfile, &networkAddressUrl, variables, id,
                    instanceArray, nbPeer, MAX_VARIABLE_NB, interval,
                    get_drone_id, pubsub_get_value, pubsub_update_variables,
                    &pubsubShouldRun);
    pubsubExited = true;

    free(instanceArray);
    JS_FreeCString(ctx, ipv6);
    JS_FreeCString(ctx, port);
    free(notConstNetIface);
    JS_FreeCString(ctx, netIface);
    return JS_NewInt32(ctx, res);
}

static JSValue js_init_pubsub(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    if (JS_ToUint32(ctx, &nbDrone, argv[0]))
        return JS_EXCEPTION;

    if (JS_ToUint32(ctx, &nbSubscriber, argv[1]))
        return JS_EXCEPTION;

    currentMessage = UA_STRING("");
    droneObjectIdList = (JSValue *) malloc((nbDrone + nbSubscriber) * sizeof(JSValueConst));
    return JS_NewInt32(ctx, 0);
}

static JSValue js_stop_pubsub(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    struct messageNode *current;

    pubsubShouldRun = false;
    while(!pubsubExited)
        sleep(1);
    free(droneObjectIdList);

    while (messageQueue.head != NULL) {
        current = messageQueue.head;
        messageQueue.head = current->next;
        delete_message_node(current);
    }
    clear_message(currentMessage);

    return JS_NewInt32(ctx, 0);
}

// Connexion management functions

static JSValue js_start(JSContext *ctx, JSValueConst thisVal,
                               int argc, JSValueConst *argv)
{
    const char *ip;
    const char *log_file;
    int port;
    int timeout;
    int res;

    ip = JS_ToCString(ctx, argv[0]);
    if (JS_ToInt32(ctx, &port, argv[1]))
        return JS_EXCEPTION;
    log_file = JS_ToCString(ctx, argv[2]);
    if (JS_ToInt32(ctx, &timeout, argv[3]))
        return JS_EXCEPTION;

    res = start(ip, port, log_file, timeout);
    JS_FreeCString(ctx, ip);
    JS_FreeCString(ctx, log_file);

    return JS_NewInt32(ctx, res);
}

static JSValue js_stop(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, stop(JS_ToBool(ctx, argv[0])));
}

static JSValue js_reboot(JSContext *ctx, JSValueConst thisVal,
                         int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, reboot());
}

// Flight state management functions

static JSValue js_arm(JSContext *ctx, JSValueConst this_val,
                             int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, arm());
}

static JSValue js_takeOff(JSContext *ctx, JSValueConst thisVal,
                          int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, takeOff());
}

static JSValue js_takeOffAndWait(JSContext *ctx, JSValueConst thisVal,
                                 int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, takeOffAndWait());
}

static JSValue js_triggerParachute(JSContext *ctx, JSValueConst thisVal,
                                   int argc, JSValueConst *argv)
{
    return JS_NewInt32(ctx, triggerParachute());
}

// Flight management functions

static JSValue js_loiter(JSContext *ctx, JSValueConst thisVal,
                         int argc, JSValueConst *argv)
{
    double la_arg_double;
    double lo_arg_double;
    double a_arg_double;
    double radius;

    if (JS_ToFloat64(ctx, &la_arg_double, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &lo_arg_double, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &a_arg_double, argv[2]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &radius, argv[3]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, loiter(la_arg_double, lo_arg_double,
                                   (float)a_arg_double, (float)radius));
}

static JSValue js_setAirspeed(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    double altitude;

    if (JS_ToFloat64(ctx, &altitude, argv[0]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setAirspeed((float)altitude));
}

static JSValue js_setTargetCoordinates(JSContext *ctx,
                                       JSValueConst thisVal,
                                       int argc, JSValueConst *argv)
{
    double la_arg_double;
    double lo_arg_double;
    double a_arg_double;

    if (JS_ToFloat64(ctx, &la_arg_double, argv[0]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &lo_arg_double, argv[1]))
        return JS_EXCEPTION;
    if (JS_ToFloat64(ctx, &a_arg_double, argv[2]))
        return JS_EXCEPTION;

    return JS_NewInt32(ctx, setTargetCoordinates(la_arg_double,
                                                 lo_arg_double,
                                                 (float)a_arg_double));
}

// Information functions

static JSValue js_getAltitude(JSContext *ctx, JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getAltitude());
}

static JSValue js_getAltitudeRel(JSContext *ctx, JSValueConst thisVal,
                                 int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getAltitudeRel());
}

static JSValue js_getInitialAltitude(JSContext *ctx,
                                     JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialAltitude());
}

static JSValue js_getInitialLatitude(JSContext *ctx,
                                     JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialLatitude());
}

static JSValue js_getInitialLongitude(JSContext *ctx,
                                      JSValueConst thisVal,
                                      int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getInitialLongitude());
}

static JSValue js_getLatitude(JSContext *ctx, JSValueConst thisVal,
                              int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getLatitude());
}

static JSValue js_getLongitude(JSContext *ctx, JSValueConst thisVal,
                               int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getLongitude());
}

static JSValue js_getTakeOffAltitude(JSContext *ctx,
                                     JSValueConst thisVal,
                                     int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getTakeOffAltitude());
}

static JSValue js_getYaw(JSContext *ctx, JSValueConst thisVal,
                                int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getYaw());
}

static JSValue js_getSpeed(JSContext *ctx, JSValueConst thisVal,
                                  int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getSpeed());
}

static JSValue js_getClimbRate(JSContext *ctx, JSValueConst thisVal,
                               int argc, JSValueConst *argv)
{
    return JS_NewFloat64(ctx, getClimbRate());
}

static JSValue js_healthAllOk(JSContext *ctx, JSValueConst this_val,
                              int argc, JSValueConst *argv)
{
    return JS_NewBool(ctx, healthAllOk());
}

static const JSCFunctionListEntry js_funcs[] = {
    JS_CFUNC_DEF("setMessage", 1, js_drone_set_message ),
    JS_CFUNC_DEF("runPubsub", 6, js_run_pubsub ),
    JS_CFUNC_DEF("stopPubsub", 0, js_stop_pubsub ),
    JS_CFUNC_DEF("start", 4, js_start ),
    JS_CFUNC_DEF("stop", 1, js_stop ),
    JS_CFUNC_DEF("reboot", 0, js_reboot ),
    JS_CFUNC_DEF("arm", 0, js_arm ),
    JS_CFUNC_DEF("takeOff", 0, js_takeOff ),
    JS_CFUNC_DEF("takeOffAndWait", 0, js_takeOffAndWait ),
    JS_CFUNC_DEF("triggerParachute", 0, js_triggerParachute ),
    JS_CFUNC_DEF("loiter", 4, js_loiter ),
    JS_CFUNC_DEF("setAirspeed", 1, js_setAirspeed ),
    JS_CFUNC_DEF("setTargetCoordinates", 3, js_setTargetCoordinates ),
    JS_CFUNC_DEF("getAltitude", 0, js_getAltitude ),
    JS_CFUNC_DEF("getAltitudeRel", 0, js_getAltitudeRel ),
    JS_CFUNC_DEF("getInitialAltitude", 0, js_getInitialAltitude ),
    JS_CFUNC_DEF("getInitialLatitude", 0, js_getInitialLatitude ),
    JS_CFUNC_DEF("getInitialLongitude", 0, js_getInitialLongitude ),
    JS_CFUNC_DEF("getLatitude", 0, js_getLatitude ),
    JS_CFUNC_DEF("getLongitude", 0, js_getLongitude ),
    JS_CFUNC_DEF("getTakeOffAltitude", 0, js_getTakeOffAltitude ),
    JS_CFUNC_DEF("getYaw", 0, js_getYaw ),
    JS_CFUNC_DEF("getAirspeed", 0, js_getSpeed ),
    JS_CFUNC_DEF("getClimbRate", 0, js_getClimbRate ),
    JS_CFUNC_DEF("healthAllOk", 0, js_healthAllOk ),
    JS_CFUNC_DEF("initPubsub", 2, js_init_pubsub ),
};

static int js_init(JSContext *ctx, JSModuleDef *m)
{
    JSValue droneProto, droneClass;

    JS_NewClassID(&jsDroneClassId);
    JS_NewClass(JS_GetRuntime(ctx), jsDroneClassId, &jsDroneClass);

    droneProto = JS_NewObject(ctx);
    JS_SetPropertyFunctionList(ctx, droneProto, js_drone_proto_funcs,
                               countof(js_drone_proto_funcs));

    droneClass = JS_NewCFunction2(ctx, js_drone_ctor, "Drone", 1,
                                   JS_CFUNC_constructor, 0);
    JS_SetConstructor(ctx, droneClass, droneProto);
    JS_SetClassProto(ctx, jsDroneClassId, droneProto);

    JS_SetModuleExport(ctx, m, "Drone", droneClass);

    return JS_SetModuleExportList(ctx, m, js_funcs,
                                  countof(js_funcs));
}

JSModuleDef *js_init_module(JSContext *ctx, const char *module_name)
{
    JSModuleDef *m;
    m = JS_NewCModule(ctx, module_name, js_init);
    if (!m)
        return NULL;
    JS_AddModuleExportList(ctx, m, js_funcs, countof(js_funcs));
    JS_AddModuleExport(ctx, m, "Drone");
    return m;
}
